"use strict";

var _errors = require("./errors.js");

var _prepackNode = require("./prepack-node.js");

var _types = require("./types.js");

var _fs = require("fs");

var _fs2 = _interopRequireDefault(_fs);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// Currently we need to explictly pass the captured variables we want to access.
// TODO: In a future version of this can be automatic.


// Prepack helper
/**
 * Copyright (c) 2017-present, Facebook, Inc.
 * All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree. An additional grant
 * of patent rights can be found in the PATENTS file in the same directory.
 */

/* eslint-disable no-shadow */

function run(Object, Array, console, JSON, process, prepackStdin, prepackFileSync, FatalError, CompatibilityValues, fs) {

  var HELP_STR = "\n    input    The name of the file to run Prepack over (for web please provide the single js bundle file)\n    --out    The name of the output file\n    --compatibility    The target environment for Prepack [" + CompatibilityValues.map(function (v) {
    return "\"" + v + "\"";
  }).join(', ') + "]\n    --mathRandomSeed    If you want Prepack to evaluate Math.random() calls, please provide a seed.\n    --srcmapIn    The input sourcemap filename. If present, Prepack will output a sourcemap that maps from the original file (pre-input sourcemap) to Prepack's output\n    --srcmapOut    The output sourcemap filename.\n    --debugNames    Changes the output of Prepack so that for named functions and variables that get emitted into Prepack's output, the original name is appended as a suffix to Prepack's generated identifier.\n    --singlePass    Perform only one serialization pass. Disables some optimizations on Prepack's output. This will speed up Prepacking but result in code with less inlining.\n    --speculate    Enable speculative initialization of modules (for the module system Prepack has builtin knowledge about). Prepack will try to execute all factory functions it is able to.\n    --trace    Traces the order of module initialization.\n    --serialize    Serializes the partially evaluated global environment as a program that recreates it. (default = true)\n    --residual    Produces the residual program that results after constant folding.\n    --profile    Enables console logging of profile information of different phases of prepack.\n  ";
  var args = Array.from(process.argv);
  args.splice(0, 2);
  var inputFilename = void 0;
  var outputFilename = void 0;
  var compatibility = void 0;
  var mathRandomSeed = void 0;
  var inputSourceMap = void 0;
  var outputSourceMap = void 0;
  var flags = {
    speculate: false,
    trace: false,
    debugNames: false,
    singlePass: false,
    logStatistics: false,
    logModules: false,
    delayUnsupportedRequires: false,
    internalDebug: false,
    serialize: false,
    residual: false,
    profile: false
  };

  while (args.length) {
    var arg = args.shift();
    if (!arg.startsWith("--")) {
      inputFilename = arg;
    } else {
      arg = arg.slice(2);
      switch (arg) {
        case "out":
          arg = args.shift();
          outputFilename = arg;
          break;
        case "compatibility":
          arg = args.shift();
          if (!CompatibilityValues.includes(arg)) {
            console.error("Unsupported compatibility: " + arg);
            process.exit(1);
          }
          compatibility = arg;
          break;
        case "mathRandomSeed":
          mathRandomSeed = args.shift();
          break;
        case "srcmapIn":
          inputSourceMap = args.shift();
          break;
        case "srcmapOut":
          outputSourceMap = args.shift();
          break;
        case "help":
          console.log("Usage: prepack.js [ --out output.js ] [ --compatibility jsc ] [ --mathRandomSeed seedvalue ] [ --srcmapIn inputMap ] [ --srcmapOut outputMap ] [ --speculate ] [ --trace ] [ -- | input.js ] [ --singlePass ] [ --debugNames ]" + "\n" + HELP_STR);
          break;
        default:
          if (arg in flags) {
            flags[arg] = true;
          } else {
            console.error("Unknown option: " + arg);
            process.exit(1);
          }
      }
    }
  }
  if (!flags.serialize && !flags.residual) flags.serialize = true;

  var resolvedOptions = Object.assign({}, {
    compatibility: compatibility,
    mathRandomSeed: mathRandomSeed,
    inputSourceMapFilename: inputSourceMap,
    sourceMaps: !!outputSourceMap
  }, flags);

  if (!inputFilename) {
    prepackStdin(resolvedOptions, processSerializedCode);
    return;
  }
  try {
    var serialized = prepackFileSync(inputFilename, resolvedOptions);
    processSerializedCode(serialized);
  } catch (x) {
    if (x instanceof FatalError) {
      // Ignore FatalError since an error has already logged
      // their errors to the console, but exit with an error code.
      process.exit(1);
    }
    console.log(x.message);
  }

  function processSerializedCode(serialized) {
    if (outputFilename) {
      console.log("Prepacked source code written to " + outputFilename + ".");
      fs.writeFileSync(outputFilename, serialized.code);
    } else {
      console.log(serialized.code);
    }
    if (outputSourceMap) {
      fs.writeFileSync(outputSourceMap, serialized.map ? JSON.stringify(serialized.map) : '');
    }
  }

  return true;
}

if (typeof __residual === 'function') {
  // If we're running inside of Prepack. This is the residual function we'll
  // want to leave untouched in the final program.
  __residual('boolean', run, Object, Array, console, JSON, process, _prepackNode.prepackStdin, _prepackNode.prepackFileSync, _errors.FatalError, _types.CompatibilityValues, _fs2.default);
} else {
  run(Object, Array, console, JSON, process, _prepackNode.prepackStdin, _prepackNode.prepackFileSync, _errors.FatalError, _types.CompatibilityValues, _fs2.default);
}
//# sourceMappingURL=prepack-cli.js.map